import os

from git import Repo, NoSuchPathError, InvalidGitRepositoryError, RemoteProgress
import logging as log
from .config import AlloConfig
from .colors import BColors
from .upgrader import AlloUpgrader

log.basicConfig(level=log.INFO)


class Progress(RemoteProgress):
    def update(self, op_code, cur_count, max_count=None, message=''):
        # TODO - make a better progress bar than text
        print('update(%s, %s, %s, %s)'%(op_code, cur_count, max_count, message))


class AlloGit:
    _repo: Repo = None

    token = None
    config: AlloConfig
    upgrader: AlloUpgrader

    def __init__(self, config: AlloConfig, token):
        self.config = config
        self.token = token
        self.init_repo()

    def init_repo(self):
        try:
            self._repo = Repo(self.config.repo_path)
            self._repo.remote().set_url("https://{}:{}@allo.dev.libriciel.fr/gitlab/{}.git".format(
                    self.config.id_client, self.token, self.config.version['gitPath']))
            self.upgrader = AlloUpgrader(self.config.repo_path)
        except (NoSuchPathError,  InvalidGitRepositoryError):
            log.info("Téléchargement du logiciel en cours...")
            os.makedirs(self.config.repo_path)
            self._repo = Repo.clone_from(
                "https://{}:{}@allo.dev.libriciel.fr/gitlab/{}.git".format(
                    self.config.id_client, self.token, self.config.version['gitPath']),
                self.config.repo_path, progress=Progress())
            log.info("We have to send event for 'installation'")
            self.upgrader = AlloUpgrader(self.config.repo_path)
            if self.upgrader.do_install() == 0:
                log.info("Version {} installée".format(self._repo.git.describe(tags=True)))
            else:
                print("Erreur d'installation")

        print(BColors.OKGREEN + "✔ {} Version {} detectée".format(self.config.code_produit, self.get_current_version()))

    def get_current_version(self):
        return self._repo.git.describe(tags=True)

    def list_versions(self):
        # Fetch new tags
        self._repo.git.fetch(tags=True)
        return self._repo.tags

    def get_versions_to_pass(self, version: str, reverse: bool = False):
        versions_list = reversed(self.list_versions()) if reverse else self.list_versions()
        versions_to_pass = []
        has_passed_current = False
        for ver in versions_list:
            if has_passed_current:
                versions_to_pass.append(str(ver))
            if str(ver) == self.get_current_version():
                has_passed_current = True
            if version == str(ver):
                break
        return versions_to_pass

    def downgrade_to(self, version: str):
        versions_to_pass = self.get_versions_to_pass(version, True)
        for version in versions_to_pass:
            self.upgrade_to_specific(version)

    def upgrade_to(self, version: str):
        versions_to_pass = self.get_versions_to_pass(version)

        for version in versions_to_pass:
            before_update_version = self.get_current_version()
            if self.upgrade_to_specific(version) != 0:
                log.error("Rollback vers la version {}".format(before_update_version))
                self.upgrade_to_specific(before_update_version)

        log.info("Upgrading to version {} complete with success".format(version))

    def upgrade_to_specific(self, version: str):
        self._repo.git.checkout("tags/{}".format(version))
        return self.upgrader.do_upgrade(version)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pickle
from pathlib import Path

import requests
from PyInquirer import prompt
from uuid import getnode
from .config import AlloConfig
from .colors import BColors
from .telem import AlloTelem
from .gitutils import AlloGit


class TestingAllo:
    config_path = str(Path.home()) + "/allo-config.dict"
    api_path = "https://allo.dev.libriciel.fr/api/client"
    config: AlloConfig
    telem: AlloTelem
    git: AlloGit

    def __init__(self, env):
        self.load_config(env)
        self.telem = AlloTelem(self.config.id_client,
                               self.config.code_produit,
                               self.config.env,
                               self.config.version,
                               self.config.teleport_token)
        self.git = AlloGit(self.config, self.get_valid_token())

        self.what_to_do()

    def save_config(self):
        with open(self.config_path, 'wb') as config_dictionary_file:
            pickle.dump(self.config, config_dictionary_file)

    def load_config(self, env):
        try:
            with open(self.config_path, 'rb') as config_dictionary_file:
                self.config: AlloConfig = pickle.load(config_dictionary_file)
        except FileNotFoundError:
            self.config = AlloConfig()
            self.config.env = env
            self.init_config_file()

    def init_config_file(self):
        error_config_msg = "Erreur de configuration, sortie du programme Allo"
        print(BColors.OKBLUE + "Merci de rentrer les informations suivantes afin d'initialiser Allo :")
        questions = [{'type': 'input', 'name': 'id_client', 'message': 'Identifiant unique Client'},
                     {'type': 'input', 'name': 'code_produit', 'message': 'Code Produit'}]
        cfg = prompt(questions)
        self.config.id_client = cfg['id_client']
        self.config.code_produit = cfg['code_produit']

        r = requests.get("{}/{}/versions".format(self.api_path, self.config.code_produit), headers={'channel': self.config.env})
        versions = r.json()
        if 'status' in versions :
            print(BColors.FAIL + "{} : {}".format(self.config.code_produit, versions['message']))
            print(BColors.FAIL + error_config_msg)
            exit(1)
        elif len(versions) is 0:
            print(BColors.FAIL + "{} : {}".format(self.config.code_produit, "Aucune versions disponibles"))
            print(BColors.FAIL + error_config_msg)
            exit(1)
        else:
            qversions = []
            for v in versions:
                qversions.append({"value": "0", "name": v['name']})
            questions = [{'type': 'list',
                          'name': 'version',
                          'message': 'Version du produit',
                          'choices': qversions}]
            answers = prompt(questions)
            self.config.version = versions[int(answers['version'])]

        if self.allo_activate():
            self.config.teleport_token = prompt(
                [{'type': 'input', 'name': 'teleport_token', 'message': 'Token de télémaintenance'}])['teleport_token']
            self.config.repo_path = prompt(
                [{'type': 'input', 'name': 'repo_path', 'message': 'Chemin d\'installation du logiciel'}])['repo_path']
            self.save_config()
        else:
            print(BColors.FAIL + "Erreur de configuration, sortie du programme Allo")
            exit(1)

    def allo_activate(self):
        r = requests.post("{}/{}/register".format(self.api_path, getnode()), json={
            "idClient": self.config.id_client,
            "product": self.config.code_produit,
            "version": self.config.version,
            "env": self.config.env
        })
        data = r.json()
        if 'status' in data:
            print(BColors.FAIL + "{} pour l'identifiant {} et le produit {}".format(data['message'],
                                                                                    self.config.id_client,
                                                                                    self.config.code_produit))
            return False
        if not data['active']:
            cfg = prompt([{'type': 'input', 'name': 'pin_code', 'message': 'Code PIN d\'activation Allo'}])
            r = requests.post("{}/{}/activate".format(self.api_path, getnode()), json={
                "idClient": self.config.id_client,
                "product": self.config.code_produit,
                "version": self.config.version,
                "env": self.config.env,
                'pin': cfg['pin_code']
            })
            data = r.json()
            if not data['active']:
                print(BColors.FAIL + "Mauvais code PIN")
                return False

        print(BColors.OKGREEN + "Connexion Allo OK")
        return True

    def get_valid_token(self):
        # Get deploy token
        r = requests.get(
            "{}/{}/{}/{}/{}/{}/token".format(self.api_path,
                                             self.config.id_client,
                                             getnode(),
                                             self.config.code_produit,
                                             self.config.version["name"],
                                             self.config.env
                                             ))
        if r.status_code is not 200:
            print(BColors.FAIL + "Impossible de récupérer le token : " + r.text)
            return False
        data = r.json()
        return data['token']

    def what_to_do(self):
        questions = [{'type': 'list', 'name': 'action', 'message': 'Que voulez-vous faire ?',
                      'choices': [
                          'Verifier la connexion',
                          'Ouvrir la télémaintenance',
                          # 'Modifier les informations de connexion',
                          'Mettre à jour',
                          'Annuler une mise à jour'
                      ]}]
        answers = prompt(questions)
        if 'action' not in answers:
            exit(0)
        if 'Verifier la connexion' in answers['action']:
            self.allo_activate()
        if 'Ouvrir la télémaintenance' in answers['action']:
            self.telem.connect()
        if 'Modifier les informations de connexion' in answers['action']:
            print("TODO")
        if 'Mettre à jour' in answers['action']:
            self.change_version(True)
        if 'Annuler une mise à jour' in answers['action']:
            self.change_version(False)
        self.what_to_do()

    def change_version(self, upgrade: bool):
        choices = self.git.get_versions_to_pass("", not upgrade)
        if len(choices) > 0:
            questions = [{'type': 'list',
                          'name': 'version',
                          'message': 'Mettre à jour vers quelle version ?' if upgrade else 'Revenir à quelle version ?',
                          'choices': choices}]
            ans = prompt(questions)
            self.git.upgrade_to(ans['version']) if upgrade else self.git.downgrade_to(ans['version'])
        else:
            print(BColors.OKBLUE +
                  ("=> Vous êtes déjà en dernière version" if upgrade else "=> Aucune version sur laquelle revenir"))


def launch():
    import sys
    env = "PROD"
    if len(sys.argv) > 1:
        env = sys.argv[1]
    TestingAllo(env)



# -*- coding: utf-8 -*-

import os
import subprocess
from pathlib import Path


class AlloTelem:
    config_file = str(Path.home()) + "/teleport-config.yml"

    def __init__(self, id_client, code_produit, env, version, token):
        if not os.path.exists(self.config_file):
            self.save_config(id_client, code_produit, token, env, version)

    def save_config(self, id_client, code_produit, token, env, version):
        teleport_yaml = ("teleport:",
                         "    nodename: {}-{}".format(id_client, code_produit),
                         '    auth_token: "{}"'.format(token),
                         '    ca_pin: "sha256:dd7a6f23f45a08ca2f80d00aa03aa81cbaf5b230614d7cf4f85f813d70d46abd"',
                         '    auth_servers:',
                         '        - allo.dev.libriciel.fr:443',
                         'auth_service:',
                         '    enabled: no',
                         'proxy_service:',
                         '    enabled: no',
                         'ssh_service:',
                         '    enabled: "yes"',
                         '    labels:',
                         '        product: {}'.format(code_produit),
                         '        version: {}'.format(version['name']),
                         '        env: {}'.format(env))

        with open(self.config_file, 'w+') as yaml_file:
            yaml_file.write('\n'.join(teleport_yaml))

    def connect(self):
        with subprocess.Popen(["sudo", "teleport", "start", '-c', self.config_file],
                              stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as proc:
            try:
                print("Télémaintenance ouverte. Pour l'interrompre, utilisez le raccourci clavier CTRL+C")
                proc.wait()
            except KeyboardInterrupt:
                print("Fin de télémaintenance")



# -*- coding: utf-8 -*-

import os
import xml.etree.ElementTree as Elt
from subprocess import call
from .xmlerror import ErrorCodes
from .colors import BColors


class AlloUpgrader:
    repo_path = None

    def __init__(self, repo_path):
        self.repo_path = repo_path

    def verify_package_xml(self):
        """
        Verify the package.xml file with packaging.xsd schema
        """
        res = os.system("xmllint --noout --schema packaging.xsd {} 2>/dev/null".format(
            self.repo_path + '/package.xml'))
        self.handle_xmllint_result_code(os.WEXITSTATUS(res))

    @staticmethod
    def run_next_cmd_as_root():
        euid = os.geteuid()
        if euid != 0:
            print("Script not started as root. Running sudo..")
            # the next line replaces the currently-running process with the sudo
            os.system('sudo' + ' echo "Logged in as root"')

    @staticmethod
    def handle_xmllint_result_code(code: int):
        default = {"message": "Unknown error.",
                   "return": ErrorCodes.UNKNOWN}
        switcher = {
            0: {"pass": True},
            1: {"message": "The package description file is not a valid XML.",
                "return": ErrorCodes.INVALID_XML},
            3: {"message": "The package description file is not valid.",
                "return": ErrorCodes.INVALID_WITH_XSD},
            127: {"messsage": "The command xmllint is not installed. Please install it and retry.",
                  "return": ErrorCodes.XMLLINT_NOT_FOUND}
        }
        handler = switcher.get(code, default)
        if not handler.get("pass"):
            print(handler.get("message"))
            exit(handler.get("return"))

    def get_root_element(self) -> Elt.Element:
        with open(self.repo_path + '/package.xml', "rb") as file:
            test = Elt.parse(file)
        return test.getroot()

    def do_upgrade(self, version):
        _rootel = self.get_root_element()

        upgrade_el = _rootel.find("Upgrades").findall("Upgrade[@Version='{}']".format(version))

        for to_upgrade in upgrade_el:
            scripts = to_upgrade.findall("Script")
            if len(scripts) > 0:
                for sh in scripts:
                    executable = "ansible-playbook --extra-vars \"repo_path={}\" ".format(self.repo_path) + \
                                 os.path.join(self.repo_path, sh.get("File"))
                    if sh.get("Sudo"):
                        self.run_next_cmd_as_root()
                        executable = "sudo " + executable
                    oscall_result = os.system(executable)
                    if os.WEXITSTATUS(oscall_result) != 0:
                        print("Error on upgrade script {}".format(sh.get("File")))
                        return os.WEXITSTATUS(oscall_result)

        print("Upgrading to version {} complete with success".format(version))
        return 0

    def do_install(self):
        # Open variables file with editor
        print(BColors.OKBLUE +
              "Allo va maintenant ouvrir le fichier de variables du logiciel. "
              "Il faut y adapter les variables requisent pour l'installation. "
              "Pressez une touche pour continuer.")
        input()

        print(BColors.OKBLUE + "Installation des dépendances")

        editor = os.environ.get('EDITOR', 'vim')
        call([editor, os.path.join(self.repo_path, "ansible/vars/default.yml")])

        # Install dependencies
        executable = "ansible-galaxy install -r {}".format(os.path.join(self.repo_path, "ansible/requirements.yml"))
        self.run_next_cmd_as_root()
        executable = "sudo " + executable
        oscall_result = os.system(executable)
        if os.WEXITSTATUS(oscall_result) != 0:
            print("Error on install ansible dependencies")
            return os.WEXITSTATUS(oscall_result)

        print(BColors.OKBLUE + "Installation du logiciel")

        # Install product
        executable = "ansible-playbook --extra-vars \"repo_path={}\" ".format(self.repo_path) + \
                     os.path.join(self.repo_path, "ansible/install.yml")
        self.run_next_cmd_as_root()
        executable = "sudo " + executable
        oscall_result = os.system(executable)
        if os.WEXITSTATUS(oscall_result) != 0:
            print("Error on ansible install script")
            return os.WEXITSTATUS(oscall_result)
